﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Problems.Programmable {
  [Item("RealEncoding", "Describes a real vector encoding.")]
  [StorableClass]
  public sealed class RealEncoding : Encoding {

    [Storable]
    private IFixedValueParameter<IntValue> lengthParameter;
    public IFixedValueParameter<IntValue> LengthParameter {
      get { return lengthParameter; }
      set {
        if (value == null) throw new ArgumentNullException("Length parameter must not be null.");
        if (lengthParameter == value) return;
        lengthParameter = value;
        OnLengthParameterChanged();
      }
    }

    [Storable]
    private IValueParameter<DoubleMatrix> boundsParameter;
    public IValueParameter<DoubleMatrix> BoundsParameter {
      get { return boundsParameter; }
      set {
        if (value == null) throw new ArgumentNullException("Bounds parameter must not be null.");
        if (boundsParameter == value) return;
        boundsParameter = value;
        OnBoundsParameterChanged();
      }
    }

    public override IEnumerable<IValueParameter> EncodingParameters {
      get { return base.EncodingParameters.Concat(new IValueParameter[] { LengthParameter, BoundsParameter }); }
    }

    public int Length {
      get { return LengthParameter.Value.Value; }
      set { LengthParameter.Value.Value = value; }
    }
    public DoubleMatrix Bounds {
      get { return BoundsParameter.Value; }
      set { BoundsParameter.Value = value; }
    }

    public IRealVectorCreator DefaultSolutionCreator {
      get {
        var creator = new UniformRandomRealVectorCreator();
        creator.RealVectorParameter.ActualName = Name;
        creator.LengthParameter.ActualName = LengthParameter.Name;
        creator.BoundsParameter.ActualName = BoundsParameter.Name;
        return creator;
      }
    }

    [StorableConstructor]
    private RealEncoding(bool deserializing) : base(deserializing) { }
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterParameterEvents();
    }

    public override IDeepCloneable Clone(Cloner cloner) { return new RealEncoding(this, cloner); }
    private RealEncoding(RealEncoding original, Cloner cloner)
      : base(original, cloner) {
      lengthParameter = cloner.Clone(original.lengthParameter);
      boundsParameter = cloner.Clone(original.boundsParameter);
      RegisterParameterEvents();
    }

    public RealEncoding(string name, int length, double min, double max)
      : base(name) {
      if (min >= max) throw new ArgumentException("min must be less than max", "min");

      var bounds = new DoubleMatrix(1, 2);
      bounds[0, 0] = min;
      bounds[0, 1] = max;

      lengthParameter = new FixedValueParameter<IntValue>(Name + "Length", new IntValue(length));
      boundsParameter = new ValueParameter<DoubleMatrix>(Name + "Bounds", bounds);
      RegisterParameterEvents();
    }

    public RealEncoding(string name, int length, IList<double> min, IList<double> max)
      : base(name) {
      if (min.Count == 0) throw new ArgumentException("Bounds must be given for the real parameters.");
      if (min.Count != max.Count) throw new ArgumentException("min must be of the same length as max", "min");
      if (min.Zip(max, (mi, ma) => mi >= ma).Any(x => x)) throw new ArgumentException("min must be less than max in each dimension", "min");

      var bounds = new DoubleMatrix(min.Count, 2);
      for (int i = 0; i < min.Count; i++) {
        bounds[i, 0] = min[i];
        bounds[i, 1] = max[i];
      }
      lengthParameter = new FixedValueParameter<IntValue>(Name + "Length", new IntValue(length));
      boundsParameter = new ValueParameter<DoubleMatrix>(Name + "Bounds", bounds);
      RegisterParameterEvents();
    }

    private void OnLengthParameterChanged() {
      RegisterLengthParameterEvents();
    }
    private void OnBoundsParameterChanged() {
      RegisterBoundsParameterEvents();
    }
    private void RegisterParameterEvents()
    {
      RegisterLengthParameterEvents();
      RegisterBoundsParameterEvents();
    }
    private void RegisterLengthParameterEvents() {
      //TODO fire correct event
      LengthParameter.ValueChanged += (o, s) => ConfigureOperators(Enumerable.Empty<IOperator>());
      LengthParameter.Value.ValueChanged += (o, s) => ConfigureOperators(Enumerable.Empty<IOperator>());
    }
    private void RegisterBoundsParameterEvents() {
      //TODO fire correct event
      BoundsParameter.ValueChanged += (o, s) => ConfigureOperators(Enumerable.Empty<IOperator>());
      boundsParameter.Value.ToStringChanged += (o, s) => ConfigureOperators(Enumerable.Empty<IOperator>());
    }

    public void DiscoverOperators() {
      var types = new List<Type>()
      {
        typeof (IRealVectorCreator),
        typeof (IRealVectorCrossover),
        typeof (IRealVectorManipulator),
        typeof (IRealVectorStdDevStrategyParameterOperator),
        typeof (IRealVectorParticleCreator),
        typeof (IRealVectorParticleUpdater),
        typeof (IRealVectorParticleUpdater),
      };
      var discoveredTypes = ApplicationManager.Manager.GetTypes(types, true, false, false);
      var operators = discoveredTypes.Select(t => (IOperator)Activator.CreateInstance(t)).ToList();
    }

    public void ConfigureOperators(IEnumerable<IOperator> operators) {
      ConfigureCreators(operators.OfType<IRealVectorCreator>());
      ConfigureCrossovers(operators.OfType<IRealVectorCrossover>());
      ConfigureManipulators(operators.OfType<IRealVectorManipulator>());
      ConfigureStdDevStrategyParameterOperators(operators.OfType<IRealVectorStdDevStrategyParameterOperator>());
      ConfigureSwarmUpdaters(operators.OfType<IRealVectorSwarmUpdater>());
      ConfigureParticleCreators(operators.OfType<IRealVectorParticleCreator>());
      ConfigureParticleUpdaters(operators.OfType<IRealVectorParticleUpdater>());
      ConfigureShakingOperators(operators.OfType<IRealVectorMultiNeighborhoodShakingOperator>());
    }

    private void ConfigureCreators(IEnumerable<IRealVectorCreator> creators) {
      foreach (var creator in creators) {
        creator.RealVectorParameter.ActualName = Name;
        creator.LengthParameter.ActualName = LengthParameter.Name;
        creator.BoundsParameter.ActualName = BoundsParameter.Name;
      }
    }
    private void ConfigureCrossovers(IEnumerable<IRealVectorCrossover> crossovers) {
      foreach (var crossover in crossovers) {
        crossover.ChildParameter.ActualName = Name;
        crossover.ParentsParameter.ActualName = Name;
        crossover.BoundsParameter.ActualName = BoundsParameter.Name;
      }
    }
    private void ConfigureManipulators(IEnumerable<IRealVectorManipulator> manipulators) {
      foreach (var manipulator in manipulators) {
        manipulator.RealVectorParameter.ActualName = Name;
        manipulator.BoundsParameter.ActualName = BoundsParameter.Name;
        manipulator.BoundsParameter.Hidden = true;
        var sm = manipulator as ISelfAdaptiveManipulator;
        if (sm != null) {
          var p = sm.StrategyParameterParameter as ILookupParameter;
          if (p != null) {
            p.ActualName = Name + "Strategy";
            p.Hidden = true;
          }
        }
      }
    }
    private void ConfigureStdDevStrategyParameterOperators(IEnumerable<IRealVectorStdDevStrategyParameterOperator> strategyOperators) {
      var bounds = (DoubleMatrix)Bounds.Clone();
      for (var i = 0; i < bounds.Rows; i++) {
        bounds[i, 1] = 0.1 * (bounds[i, 1] - bounds[i, 0]);
        bounds[i, 0] = 0;
      }
      foreach (var s in strategyOperators) {
        var c = s as IRealVectorStdDevStrategyParameterCreator;
        if (c != null) {
          c.BoundsParameter.Value = (DoubleMatrix)bounds.Clone();
          c.LengthParameter.ActualName = LengthParameter.Name;
          c.StrategyParameterParameter.ActualName = Name + "Strategy";
        }
        var m = s as IRealVectorStdDevStrategyParameterManipulator;
        if (m != null) {
          m.BoundsParameter.Value = (DoubleMatrix)bounds.Clone();
          m.StrategyParameterParameter.ActualName = Name + "Strategy";
        }
        var mm = s as StdDevStrategyVectorManipulator;
        if (mm != null) {
          mm.GeneralLearningRateParameter.Value = new DoubleValue(1.0 / Math.Sqrt(2 * Length));
          mm.LearningRateParameter.Value = new DoubleValue(1.0 / Math.Sqrt(2 * Math.Sqrt(Length)));
        }
        var x = s as IRealVectorStdDevStrategyParameterCrossover;
        if (x != null) {
          x.ParentsParameter.ActualName = Name + "Strategy";
          x.StrategyParameterParameter.ActualName = Name + "Strategy";
        }
      }
    }
    private void ConfigureSwarmUpdaters(IEnumerable<IRealVectorSwarmUpdater> swarmUpdaters) {
      foreach (var su in swarmUpdaters) {
        su.RealVectorParameter.ActualName = Name;
      }
    }
    private void ConfigureParticleCreators(IEnumerable<IRealVectorParticleCreator> particleCreators) {
      foreach (var particleCreator in particleCreators) {
        particleCreator.RealVectorParameter.ActualName = Name;
        particleCreator.BoundsParameter.ActualName = BoundsParameter.Name;
        particleCreator.ProblemSizeParameter.ActualName = LengthParameter.Name;
      }
    }
    private void ConfigureParticleUpdaters(IEnumerable<IRealVectorParticleUpdater> particleUpdaters) {
      foreach (var particleUpdater in particleUpdaters) {
        particleUpdater.RealVectorParameter.ActualName = Name;
        particleUpdater.BoundsParameter.ActualName = BoundsParameter.Name;
      }
    }
    private void ConfigureShakingOperators(IEnumerable<IRealVectorMultiNeighborhoodShakingOperator> shakingOperators) {
      foreach (var shakingOperator in shakingOperators) {
        shakingOperator.RealVectorParameter.ActualName = Name;
      }
    }
  }
}